import { BufferWriter, countItems } from './src/utils';
import { Base64 } from 'js-base64';

export default class YGOProDeck {
  main: number[] = [];
  extra: number[] = [];
  side: number[] = [];

  bufferLength() {
    const counted = [this.main, this.extra, this.side].map(countItems);
    return counted.reduce((a, b) => a + b.size * 4, 0);
  }

  toUint8Array() {
    const counted = [this.main, this.extra, this.side].map(countItems);
    const writer = new BufferWriter(
      counted.reduce((a, b) => a + b.size * 4, 0),
    );
    const writeCards = (countMap: Map<number, number>, type: number) => {
      // each card: 28 bits for id, 2 bits(0, 1, 2, 3) for type(0: main, 1: extra, 2: side, 3: unknown), 2 bits for count (0: 1, 1: 2, 2: 3, 3: 4)
      for (const [id, count] of countMap.entries()) {
        if (count > 4) {
          throw new Error(`Too many cards: ${id}`);
        }
        const value = (id & 0xfffffff) | (type << 28) | ((count - 1) << 30);
        writer.writeUint32LE(value);
      }
    };
    counted.forEach(writeCards);
    return writer.buffer;
  }

  toEncodedString() {
    return Base64.fromUint8Array(this.toUint8Array(), true);
  }

  toString() {
    return this.toEncodedString();
  }

  fromUint8Array(buf: Uint8Array) {
    for (let i = 0; i < buf.length - 3; i += 4) {
      const value =
        buf[i] | (buf[i + 1] << 8) | (buf[i + 2] << 16) | (buf[i + 3] << 24);
      const id = value & 0xfffffff;
      const type = (value >> 28) & 0x3;
      const count = ((value >> 30) & 0x3) + 1;
      const cards = [this.main, this.extra, this.side][type];
      for (let j = 0; j < count; j++) {
        cards.push(id);
      }
    }
    return this;
  }

  static fromUint8Array(buf: Uint8Array) {
    return new YGOProDeck().fromUint8Array(buf);
  }

  fromEncodedString(str: string) {
    return this.fromUint8Array(Base64.toUint8Array(str));
  }

  static fromEncodedString(str: string) {
    return new YGOProDeck().fromEncodedString(str);
  }

  toYdkString() {
    return [
      '#created by ygopro-deck-encode',
      '#main',
      ...this.main.map((id) => id.toString()),
      '#extra',
      ...this.extra.map((id) => id.toString()),
      '!side',
      ...this.side.map((id) => id.toString()),
    ].join('\n');
  }

  fromYdkString(str: string) {
    const lines = str.split('\n');
    let current = this.main;
    for (const _line of lines) {
      const line = _line.trim();
      if (line === '#main') {
        current = this.main;
      }
      if (line === '#extra') {
        current = this.extra;
      }
      if (line === '!side') {
        current = this.side;
      }
      if (line.match(/^\d+$/)) {
        current.push(parseInt(line, 10));
      }
    }
    return this;
  }

  static fromYdkString(str: string) {
    return new YGOProDeck().fromYdkString(str);
  }

  toUpdateDeckPayload() {
    const cards = [...this.main, ...this.extra, ...this.side];
    const writer = new BufferWriter(cards.length * 4 + 8)
      .writeUint32LE(this.main.length + this.extra.length)
      .writeUint32LE(this.side.length);
    cards.forEach((id) => writer.writeUint32LE(id));
    return writer.buffer;
  }
}
